// gpio.h 2013-01-30 20:20
//
// (c) 2013 Markus Weber, Nuernberg
// GNU GPL v2 (see Readme.txt)
//
// This source code file adds general purpose i/o functions
// to USBasp version "guloprog" for ATtiny85 microcontroller.
//
// Include this file in USBasp's main.c directly before function
// definition of usbFunctionSetup() and place three function calls
// at these positions:
// if(data[1]>='A' && data[1]<='D') len= gpio_control(data); else
//                      in function usbFunctionSetup(), right before
//                      this line: if (data[1] == USBASP_FUNC_CONNECT)
// gpio_ini():          in main(), before sei();
// gpio_main():         in main(), right after usbPoll();

#include <inttypes.h>

#define GPIO_A 4  // i/o pin A: PB4 (pin 3, no resistor)
#define GPIO_B 3  // i/o pin B: PB3 (pin 2)
#define GPIO_C 5  // i/o pin C: PB5 (pin 1, weak output)
#define GPIO_D 0  // i/o pin D: PB0 (pin 5, no ADC)
#define GPIO_ADMUX ((0<<REFS2)|(1<<REFS1)|(0<<REFS0))
int8_t gpio_val[6];  // new value (0..100, -1, -2)
uint8_t gpio_adc_low[6];  // ADC result, low byte
uint8_t gpio_adc_high[6];  // ADC result, high byte
uint8_t gpio_tcnt;  // old counter value of Timer 1
int8_t gpio_old_val,gpio_new_val;
  // for synchronisation of the procedures gpio_control() and gpio_min()



static inline uint8_t gpio_control(const uint8_t* data) {
  // process a USB control message which has been received from host;
  // data[1]: i/o pin ('A'..'D'), must be in this range;
  // data[2]: control value for this pin;
  //          0: set pin to Low;
  //          1..99: activate PWM for this pin, 1..99 % High;
  //          100: set pin to High;
  //          -1: switch pin to input;
  //          -2: switch pin to input, activate pull-up resistor;
  //          -6: just get input values for this pin;
  // return: 4;
  // in case data[2]==-6, the replyBuffer[] is set to these values:
  // replyBuffer[0]: 0: pin is Low; 1: pin is High;
  // replyBuffer[1]: 0;
  // replyBuffer[2]: ADC value of the requested pin, low byte
  // replyBuffer[3]: ADC value of the requested pin, high byte
  uint8_t iopin;
  int8_t newval;

  iopin= data[1]-'A';
  newval= data[2];
  if(newval==-6) {  // no output, just get input values
    if(iopin==0) {
      replyBuffer[0]= PINB & (1<<GPIO_A);
      replyBuffer[2]= gpio_adc_low[GPIO_A];
      replyBuffer[3]= gpio_adc_high[GPIO_A];
      }
    else if(--iopin==0) {
      replyBuffer[0]= PINB & (1<<GPIO_B);
      replyBuffer[2]= gpio_adc_low[GPIO_B];
      replyBuffer[3]= gpio_adc_high[GPIO_B];
      }
    else if(--iopin==0) {
      replyBuffer[0]= PINB & (1<<GPIO_C);
      replyBuffer[2]= gpio_adc_low[GPIO_C];
      replyBuffer[3]= gpio_adc_high[GPIO_C];
      }
    else {
      replyBuffer[0]= PINB & (1<<GPIO_D);
      // fake ADC values since i/o pin D hast no ADC
      if(replyBuffer[0]==0) {  // Low
        replyBuffer[2]= 0;
        replyBuffer[3]= 0;
        }
      else {  // High
        replyBuffer[2]= 0xff;
        replyBuffer[3]= 3;
        }
      }
    if(replyBuffer[0]!=0) replyBuffer[0]= 1;
    replyBuffer[1]= 0;
    }  // no output, just get input values
  else {  // store switch command
    if(iopin==0)
      gpio_val[GPIO_A]= newval;
    else if(--iopin==0)
      gpio_val[GPIO_B]= newval;
    else if(--iopin==0)
      gpio_val[GPIO_C]= newval;
    else
      gpio_val[GPIO_D]= newval;
    gpio_new_val++;  // inform gpio_main() about the change
    }  // store switch command
  return 4;
  }  // gpio_control()



static inline void gpio_ini() {
  // initialize general purpose i/o funtions;
  #define D(x) gpio_val[x]= -2; gpio_adc_low[x]= 0; gpio_adc_high[x]= 0;
  D(GPIO_A) D(GPIO_B) D(GPIO_C) D(GPIO_D)
  #undef D
  gpio_old_val= gpio_new_val= 0;
  gpio_tcnt= 0;
  // initialize timer
  TCNT1= 0;  // Timer 1 starts counting at 0
  OCR1C= 99;  // maximum value for Timer 1
  TCCR1= (1<<CTC1)|(1<<CS13)|(0<<CS12)|(1<<CS11)|(1<<CS10);
    // start Timer 1 with prescaler 1024; -> 16500 Hz
    // set CTC1 (Clear Timer/Counter on Compare Match)
  // initialize ADC
  ADMUX= GPIO_ADMUX|(1<<MUX1)|(0<<MUX0);
    // use internal 1.1 V reference, start with i/o pin A (PB4, pin 3)
  ADCSRA= (1<<ADEN)|(1<<ADPS2)|(1<<ADPS1)|(1<<ADPS0);
    // activate ADC, set ADC prescaler to 128
  }  // gpio_ini()



static inline void gpio_main() {

  // process timer step
  if(gpio_tcnt!=TCNT1) {  // timer counter did one step forward
    int8_t val;

    gpio_tcnt++;

    // care about PWM output (soft PWM)
    #define D(x) val= gpio_val[x]; \
      if(val>0 && val<100) {  /* PWM output requested */ \
        if(gpio_tcnt<val)  /* we are in High range */ \
          PORTB|= (1<<x); \
        else  /* we are in Low range */ \
          PORTB&= ~(1<<x); \
        }  // PWM output requested
    D(GPIO_A) D(GPIO_B) D(GPIO_C) D(GPIO_D)
    #undef D

    usbPoll();

    // care about ADC results
    if(gpio_tcnt==3) {  // get ADC data for C and start conversion for A
      gpio_adc_low[GPIO_C]= ADCL;
      gpio_adc_high[GPIO_C]= ADCH;
      ADMUX= GPIO_ADMUX|(1<<MUX1)|(0<<MUX0);
      ADCSRA|= (1<<ADSC);
      }
    if(gpio_tcnt==36) {  // get ADC data for A and start conversion for B
      gpio_adc_low[GPIO_A]= ADCL;
      gpio_adc_high[GPIO_A]= ADCH;
      ADMUX= GPIO_ADMUX|(1<<MUX1)|(1<<MUX0);
      ADCSRA|= (1<<ADSC);
      }
    if(gpio_tcnt==69) {  // get ADC data for B and start conversion for C
      gpio_adc_low[GPIO_B]= ADCL;
      gpio_adc_high[GPIO_B]= ADCH;
      ADMUX= GPIO_ADMUX|(0<<MUX1)|(0<<MUX0);
      ADCSRA|= (1<<ADSC);
      }
    }  // timer counter did one step forward

  usbPoll();

  // process user request for changes of i/o pin states
  if(gpio_new_val!=gpio_old_val) {  // at least one value has been changed
    int8_t val;

    gpio_old_val++;  // remember that we have reacted
    // set new values to i/o pins;
    // gpio_val[]: values which represent the i/o pin states;
    //             0: Low; 1: High; 1..9: PWM;
    //             -1: input; -2: input + pull-up;

    #define D(x) val= gpio_val[x]; \
    if(val<0) {  /* switch pin to input, set pull-up if requested */ \
      DDRB&= ~(1<<x); \
      if(++val==0) PORTB&= ~(1<<x); \
      else PORTB|= (1<<x); \
      } \
    else {  /* switch to output */ \
      if(val==0)  /* switch to Low */ \
        PORTB&= ~(1<<x); \
      else if(val==100)  /* switch to High */ \
        PORTB|= (1<<x); \
      /* else PWM, do not switch, just wait for PWM cycle */ \
      DDRB|= (1<<x);  /* switch direction to output */ \
      }  // switch to output
    D(GPIO_A) D(GPIO_B) D(GPIO_C) D(GPIO_D)
    #undef D
    }  // at least one value has been changed
  }  // gpio_main()

